"""SenseME Discovery.

This class periodically broadcasts discovery packets and listens for response messages
from SenseME fan by Big Ass Fans. Discovered fans are announced via callbacks.

Based on work from Bruce at http://bruce.pennypacker.org/tag/senseme-plugin/
and https://github.com/bpennypacker/SenseME-Indigo-Plugin

Based on work from TomFaulkner at https://github.com/TomFaulkner/SenseMe

Source can be found at https://github.com/mikelawrence/aiosenseme
"""
import asyncio
import inspect
import ipaddress
import logging
import random
import socket
import time
import traceback

import ifaddr

from .fan import SensemeFan

_LOGGER = logging.getLogger(__name__)

PORT = 31415


class SensemeDiscoveryEndpoint:
    """High-level endpoint for SenseME Discovery protocol."""

    receive_queue = asyncio.Queue()

    def __init__(self, ip: str = None):
        """Initialize Senseme Discovery Endpoint."""
        self.opened = False
        self.transport = None
        self.ip = ip

    def abort(self):
        """Close the transport immediately.

        Buffered write data will be lost.
        """
        self.opened = False
        if self.transport is None:
            return
        self.transport.abort()
        self.close()

    def close(self):
        """Close the transport gracefully.

        Buffered write data will be sent.
        """
        self.opened = False
        if self.transport is None:
            return
        self.receive_queue.put_nowait(None)  # tell receive() socket is closed
        if self.transport:
            self.transport.close()

    def is_closing(self) -> bool:
        """Return True if the endpoint is closed or closing."""
        if not self.opened:
            return True  # unopened connection is closed
        if self.transport is None:
            return True  # opened connection but no transport is closed
        return self.transport.is_closing()

    async def receive(self) -> SensemeFan:
        """Wait for discovered device and return it.

        Return None when the socket is closed.
        This method is a coroutine.
        """
        while True:
            if self.receive_queue.empty() and self.transport.is_closing():
                return None
            rsp = await self.receive_queue.get()
            if rsp is None:
                return None
            msg = rsp[0]
            addr = rsp[1]
            if len(msg) > 200 or len(msg) < 31:
                continue
            if msg[0] != "(" or msg[-1] != ")":
                continue
            msg = msg[1:-1]
            msg_data = msg.split(";")
            if len(msg_data) != 5:
                continue
            if msg_data[4] == "SWITCH,SENSEME":
                _LOGGER.debug("Ignored Wall Switch '%s' on %s", msg, self.ip)
                continue
            _LOGGER.debug("Received '%s' from %s on %s", msg, addr, self.ip)
            device = SensemeFan(msg_data[0], msg_data[3], addr, msg_data[4])
            return device

    def send_broadcast(self):
        """Send the SenseME Discovery broadcast packet."""
        if not self.is_closing():
            data = "<ALL;DEVICE;ID;GET>".encode("utf-8")
            self.transport.sendto(data, ("<broadcast>", PORT))
            _LOGGER.debug("Discovery broadcast on %s", self.ip)


class SensemeDiscoveryProtocol(asyncio.DatagramProtocol):
    """Datagram protocol for SenseME Discovery."""

    def __init__(self, endpoint: SensemeDiscoveryEndpoint):
        """Initialize Senseme Discovery Protocol."""
        self._endpoint = endpoint

    # Protocol methods
    def connection_made(self, transport: asyncio.Protocol):
        """Socket connect on SenseME Discovery Protocol."""
        self._endpoint.transport = transport
        self._endpoint.opened = True
        _LOGGER.debug("Listening on %s", self._endpoint.ip)
        self._endpoint.send_broadcast()

    def connection_lost(self, exc):  # pylint: disable=unused-argument
        """Lost connection SenseME Discovery Protocol."""
        _LOGGER.debug("Listener closed on %s", self._endpoint.ip)
        self._endpoint.close()  # half-closed connections are not permitted

    # Datagram protocol methods
    def datagram_received(self, data: str, addr: str):
        """UDP packet received on SenseME Discovery Protocol."""
        if data:
            msg = data.decode("utf-8")
            try:
                self._endpoint.receive_queue.put_nowait((msg, addr[0]))
            except asyncio.QueueFull:
                _LOGGER.error("Receive queue full")

    def error_received(self, exc):
        """Error on SenseME Discovery Protocol."""
        _LOGGER.debug(
            "Protocol error on %s\n%s", self._endpoint.ip, traceback.format_exc()
        )
        self._endpoint.close()


class SensemeDiscovery:
    """SenseME Discovery Class.

    This class periodically broadcasts discovery packets and listens
    for response messages from SenseME fans by Big Ass Fans.
    """

    _devices = []  # all SensemeDiscovery objects use the same device list

    def __init__(self, start_first: bool = True, refresh_minutes: int = 5):
        """Initialize Senseme Discovery Protocol."""
        self.start_first = start_first
        self.refresh_minutes = refresh_minutes
        self._is_running = False
        self._callbacks = []
        self._broadcaster_task = None

    @property
    def devices(self):
        """Get the current list of discovered fans."""
        return self._devices

    def add_callback(self, callback):
        """Add callback function/coroutine.

        Called when parameters are updated.
        """
        if callback not in self._callbacks:
            self._callbacks.append(callback)
            if inspect.iscoroutinefunction(callback):
                asyncio.get_event_loop().create_task(callback(self._devices.copy()))
                _LOGGER.debug("Added coroutine callback")
            else:
                callback(self._devices.copy())
                _LOGGER.debug("Added function callback")

    def remove_callback(self, callback):
        """Remove existing callback function/coroutine."""
        if callback in self._callbacks:
            self._callbacks.remove(callback)
            _LOGGER.debug("Removed callback")

    async def _create_endpoints(self):
        """Create an endpoint per interface and return as a list of endpoints."""
        endpoints = []
        loop = asyncio.get_event_loop()
        listening = 0
        for adapter in ifaddr.get_adapters():
            for ip in adapter.ips:
                if ip.is_IPv4 and not ipaddress.ip_address(ip.ip).is_loopback:
                    endpoint = SensemeDiscoveryEndpoint(ip.ip)
                    # _LOGGER.debug("Found IPv4 %s", ip.ip)
                    try:
                        await loop.create_datagram_endpoint(
                            lambda ep=endpoint: SensemeDiscoveryProtocol(ep),
                            local_addr=(ip.ip, PORT),
                            family=socket.AF_INET,
                            allow_broadcast=True,
                        )
                        endpoints.append(endpoint)
                        listening += 1
                    except OSError:
                        last_error = (
                            f"Create datagram endpoint error on {ip.ip}\n"
                            f"{traceback.format_exc()}"
                        )
                        _LOGGER.debug(last_error)
        if listening == 0:
            # failed to bind to any address
            _LOGGER.error(
                "Failed to listen for discovery responses on any address. "
                "Last error\n%s",
                last_error,
            )

        return endpoints

    async def _broadcaster(self):
        """Periodically broadcast discovery packet.

        If the underlying socket has an error this task will exit.
        This method is a coroutine.
        """
        self._is_running = True
        loop = asyncio.get_event_loop()
        device = None
        while True:
            try:
                found_new = 0
                found_total = 0
                found_devices = []
                endpoints = await self._create_endpoints()
                start = time.time()
                while True:
                    try:
                        device = await asyncio.wait_for(endpoints[0].receive(), 1)
                    except asyncio.TimeoutError:
                        device = None
                        if time.time() - start < 5:
                            for endpoint in endpoints:
                                endpoint.send_broadcast()
                        else:
                            for endpoint in endpoints:
                                endpoint.abort()
                            endpoints = []
                            break
                    if device is not None:
                        if device not in self._devices:
                            if self.start_first:
                                if await device.update():
                                    self._devices.append(device)
                                    _LOGGER.debug("Discovered %s", device)
                                    found_new += 1
                                else:
                                    _LOGGER.debug("Failed to start %s", device.name)
                            else:
                                if await device.fill_out_secondary_info():
                                    self._devices.append(device)
                                    _LOGGER.debug("Discovered %s", device)
                                    found_new += 1
                                else:
                                    _LOGGER.debug(
                                        "Failed to retrieve secondary info for %s",
                                        device.name,
                                    )
                        if device not in found_devices:
                            found_devices.append(device)
                            found_total += 1
                        for callback in self._callbacks:
                            if inspect.iscoroutinefunction(callback):
                                loop.create_task(callback(self._devices.copy()))
                            else:
                                callback(self._devices.copy())
                found_old = found_total - found_new
                found_devices = None
                if found_old > 1:
                    _LOGGER.debug(
                        "Discovered %s existing fan%s",
                        found_old,
                        "" if found_old == 1 else "s",
                    )
                elif found_new == 0:
                    _LOGGER.debug("Discovered 0 fans")
                await asyncio.sleep(self.refresh_minutes * 60 + random.uniform(-10, 10))
            except asyncio.CancelledError:
                _LOGGER.debug("Broadcaster task cancelled")
                return
            except Exception:
                _LOGGER.error("Broadcaster task error\n%s", traceback.format_exc())
                raise
            finally:
                for endpoint in endpoints:
                    endpoint.abort()
                endpoints = None
                found_devices = None
        _LOGGER.error("Broadcaster task ended")

    def start(self):
        """Start both broadcaster and listener tasks.

        Will maintain a list of discovered fans.
        """
        if not self._is_running:
            loop = asyncio.get_event_loop()
            self._broadcaster_task = loop.create_task(self._broadcaster())

    def stop(self):
        """Stop both broadcaster and listener tasks.

        Any discovered fans will remain in memory and will continue to update.
        """
        if self._is_running is True:
            self._broadcaster_task.cancel()
            self._is_running = False

    def remove_discovered_devices(self):
        """Stop both broadcaster and listener tasks.

        Any discovered fans will be stopped and removed from memory.
        """
        self.stop()
        for fan in self._devices:
            fan.stop()
        self._devices = []


async def discover_any(timeout_seconds=5) -> bool:
    """Return True if any SenseME fans are found on the network.

    This function will always take timeout_seconds to complete.
    This method is a coroutine.
    """
    discovery = SensemeDiscovery(True, 1)
    discovery.start()
    await asyncio.sleep(timeout_seconds)
    count = len(discovery.devices)
    discovery.stop()
    _LOGGER.debug("Discovered %s fan%s", count, "" if count == 1 else "s")
    return count > 0


async def discover(value, timeout_seconds=5) -> SensemeFan:
    """Discover a fan with a fan name, room name or IP Address.

    None is returned if the fan was not found.
    This function will take up timeout_seconds to complete.
    This method is a coroutine.
    """
    start = time.time()
    discovery = SensemeDiscovery(True, 1)
    try:
        discovery.start()
        while True:
            await asyncio.sleep(0.1)
            devices = discovery.devices.copy()
            for device in devices:
                if device == value:
                    return device
            if time.time() - start > timeout_seconds:
                return None
    finally:
        discovery.stop()



"""Script interface for aiosenseme library."""

import argparse
import asyncio
import logging
from typing import List

import aiosenseme
from aiosenseme import SensemeDiscovery, SensemeFan
from aiosenseme import __version__

ARGS = argparse.ArgumentParser(
    description="Discover and control Haiku/SenseME fans by Big Ass Fans."
)
ARGS.add_argument(
    "-V",
    "--version",
    action="store_true",
    dest="version",
    default=False,
    help="display version number",
)
ARGS.add_argument(
    "--debug",
    action="store_true",
    dest="debug",
    default=False,
    help="enable debug level logging",
)
ARGS.add_argument(
    "-d",
    "--discover",
    action="store_true",
    dest="discover",
    default=False,
    help="discover all fans on the network",
)
ARGS.add_argument(
    "-m",
    "--models",
    action="store_true",
    dest="models",
    default=False,
    help="list known fan models",
)
ARGS.add_argument(
    "-n",
    "--name",
    action="store",
    dest="name",
    default=None,
    help="fan name, room name or IP address",
)
ARGS.add_argument(
    "-p",
    "--fan",
    action="store",
    dest="fan",
    default=None,
    choices=["on", "off"],
    help="fan power",
)
ARGS.add_argument(
    "-s",
    "--speed",
    action="store",
    dest="speed",
    default=None,
    type=int,
    choices=range(0, 8),
    help="fan speed",
)
ARGS.add_argument(
    "-l",
    "--light",
    action="store",
    dest="light",
    default=None,
    choices=["on", "off"],
    help="light power",
)
ARGS.add_argument(
    "-b",
    "--brightness",
    action="store",
    dest="brightness",
    default=None,
    type=int,
    choices=range(0, 17),
    help="light brightness",
)
ARGS.add_argument(
    "-w",
    "--whoosh",
    action="store",
    dest="whoosh",
    default=None,
    choices=["on", "off"],
    help="whoosh mode",
)

# array of discovered devices
_DEVICES = []


def print_fan(fan: SensemeFan):
    """Print information about fan."""
    msg = f"{fan.name}\n"
    if fan.room_status:
        msg += f"  Room Name: {fan.room_name}, Room Type: {fan.room_type}\n"
    if fan.has_light:
        msg += f"  Model: {fan.model} with light, "
    else:
        msg += f"  Model: {fan.model} without light, "
    msg += f"FW Version: {fan.fw_version}\n"
    msg += f"  IP Addr: {fan.ip}, MAC Addr: {fan.mac}\n"
    msg += f"  Token: {fan.network_token}"
    print(msg)


def print_fan_state(prefix: str, fan: SensemeFan):
    """Print information about fan current state."""
    msg = prefix
    if fan.fan_on:
        msg += f": Fan is on (speed: {fan.fan_speed})"
    else:
        msg += ": Fan is off"
    if fan.light_on:
        msg += f", Light is on (brightness: {fan.light_brightness})"
    else:
        msg += ", Light is off"
    if fan.fan_whoosh:
        msg += ", Whoosh: on"
    else:
        msg += ", Whoosh: off"
    print(msg)


async def discovered(fans: List[SensemeFan]):
    """Discovered fan callback function.

    Called when discovery has detected a SenseME fan.
    Each time a device is discovered all devices discovered are reported.
    """
    global _DEVICES
    for fan in fans:
        if fan not in _DEVICES:
            _DEVICES.append(fan)
            print_fan(fan)


async def process_args():
    """Process command line arguments."""
    args = ARGS.parse_args()
    if args.debug is True:
        logging.basicConfig(level=logging.DEBUG)
    if args.version is True:
        print("aiosenseme %s" % __version__)
        print("Copyright (C) 2020 by Mike Lawrence")
        print(
            "This is free software. You may redistribute copies of it under the terms"
        )
        print(
            "of the GNU General Public License <http://www.gnu.org/licenses/gpl.html>."
        )
        print("There is NO WARRANTY, to the extent permitted by law.")
    elif args.discover is True:
        try:
            discovery = SensemeDiscovery(True, 1)
            discovery.add_callback(discovered)
            discovery.start()
            await asyncio.sleep(4)
        finally:
            discovery.stop()
    elif args.models is True:
        msg = "Known fan models: "
        first = True
        for model in SensemeFan.models():
            if first:
                first = False
                msg += model
            else:
                msg += ", " + model
        print(msg)
        return
    else:
        if args.name is None:
            print("You must specify a fan name using -n or --name")
            return
        fan = await aiosenseme.discover(args.name, 2)
        if fan is None:
            print(f"Fan/Room/IP address '{args.name}' not found")
            return
        print_fan(fan)
        print_fan_state("State", fan)
        try:
            changed = False
            if args.whoosh is not None:
                print(f"whoosh={args.whoosh}")
                if fan.fan_whoosh != (args.whoosh == "on"):
                    changed = True
                fan.fan_whoosh = args.whoosh == "on"
            if args.speed is not None:
                if args.fan is not None:
                    print(
                        "When specifying --fanspeed there is no " "reason to set --fan"
                    )
                if fan.fan_speed != args.speed:
                    changed = True
                fan.fan_speed = args.speed
            elif args.fan is not None:
                if fan.fan_on != (args.fan == "on"):
                    changed = True
                fan.fan_on = args.fan == "on"
            if args.brightness is not None:
                if args.light is not None:
                    print(
                        "When specifying --brightness there is no "
                        "reason to set --light"
                    )
                if fan.light_brightness != args.brightness:
                    changed = True
                fan.light_brightness = args.brightness
            elif args.light is not None:
                if fan.light_on != (args.light == "on"):
                    changed = True
                fan.light_on = args.light == "on"
            if changed:
                await asyncio.sleep(0.5)
                print_fan_state("New State", fan)

        finally:
            fan.stop()


def cli():
    """Command line interface for SensemeDiscovery."""
    task = asyncio.Task(process_args())
    loop = asyncio.get_event_loop()
    loop.run_until_complete(task)



"""aiosenseme library."""

from .discovery import discover
from .discovery import discover_any
from .discovery import SensemeDiscovery
from .fan import SensemeFan
from .version import __version__

__all__ = [
    "SensemeFan",
    "SensemeDiscovery",
    "discover_any",
    "discover",
]

# flake8: noqa



#!/usr/bin/env python
"""Setup aiosenseme library."""
from __future__ import print_function

import sys

from setuptools import find_packages, setup

__version__ = "Unknown"
exec(open("aiosenseme/version.py").read())

if sys.version_info < (3, 6):
    error = """
    aiosenseme supports Python 3.6 and above.

    Python {py} detected.

    Please install using pip3 on Python 3.6 or above.
    """.format(
        py=".".join([str(v) for v in sys.version_info[:3]])
    )

    print(error, file=sys.stderr)
    sys.exit(1)

setup(
    name="aiosenseme",
    version=__version__,
    description="SenseME by Big Ass Fans asynchronous Python library",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="mikelawrence",
    author_email="mikealawr@gmail.com",
    url="https://github.com/mikelawrence/aiosenseme",
    packages=find_packages(),
    include_package_data=True,
    license="GPL3",
    install_requires=["ifaddr>=0.1.6"],
    entry_points={
        "console_scripts": ["aiosenseme = aiosenseme.scripts.commandline:cli"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Framework :: AsyncIO",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Home Automation",
    ],
    keywords="Haiku HaikuHome SenseME fan home automation BigAssFans",
    python_requires=">=3.6",
)



"""SensemeFan Class.

This class connects to a SenseME fan by Big Ass Fans to retrieve and maintain its
complete state. It keep a connection always open to the fan and will receive virtually
instantaneous state changes (push) from any source (i.e. remote control or Haiku App).
The always open connection will automatically reconnect on errors or when the SenseME
fan disconnects. Control and access to the fan's current state is provided by class
properties. State changes are announced via callbacks.

Based on work from Bruce at http://bruce.pennypacker.org/tag/senseme-plugin/
and https://github.com/bpennypacker/SenseME-Indigo-Plugin

Based on work from TomFaulkner at https://github.com/TomFaulkner/SenseMe

Source can be found at https://github.com/mikelawrence/aiosenseme
"""
import asyncio
import inspect
import ipaddress
import logging
import random
import time
import traceback
from typing import Any, Callable, Tuple

_LOGGER = logging.getLogger(__name__)

PORT = 31415

ONOFF = ["ON", "OFF"]
DIRECTIONS = ["FWD", "REV"]
AUTOCOMFORTS = ["OFF", "COOLING", "HEATING", "FOLLOWTSTAT"]
ROOM_TYPES = [
    "Undefined",  # 0, not in a room
    "Other",  # 1
    "Master Bedroom",  # 2
    "Bedroom",  # 3
    "Den",  # 4
    "Family Room",  # 5
    "Living Room",  # 6
    "Kids Room",  # 7
    "Kitchen",  # 8
    "Dining Room",  # 9
    "Basement",  # 10
    "Office",  # 11
    "Patio",  # 12
    "Porch",  # 13
    "Hallway",  # 14
    "Entryway",  # 15
    "Bathroom",  # 16
    "Laundry",  # 17
    "Stairs",  # 18
    "Closet",  # 19
    "Sunroom",  # 20
    "Media Room",  # 21
    "Gym",  # 22
    "Garage",  # 23
    "Outside",  # 24
    "Loft",  # 25
    "Playroom",  # 26
    "Pantry",  # 27
    "Mudroom",  # 28
]

FAN_MODEL_TYPES = {
    "FAN,HAIKU,SENSEME": "Haiku Fan",
    "FAN,HAIKU,HSERIES": "Haiku Fan",  # H Series is now called plain Haiku
    "FAN,LSERIES": "Haiku L Fan",
}


class SensemeEndpoint:
    """High-level endpoint for SenseME protocol."""

    def __init__(self):
        """Initialize Senseme Discovery Endpoint."""
        self.receive_queue = asyncio.Queue()
        self.opened = False
        self.transport = None

    def abort(self):
        """Close the transport immediately. Buffered write data will be flushed."""
        self.transport.abort()
        self.close()

    def close(self):
        """Close the transport gracefully. Buffered write data will be sent."""
        if self.transport is None:
            return
        self.receive_queue.put_nowait(None)  # tell receive() socket is closed
        if self.transport:
            self.transport.close()

    def is_closing(self) -> bool:
        """Return True if the endpoint is closed or closing."""
        if not self.opened:
            return False  # unopened connection is not closed
        if self.transport is None:
            return True  # opened connection but no transport is closed
        return self.transport.is_closing()

    async def receive(self) -> str:
        """Wait for a message from the SenseME fan.

        Return None when the socket is closed.
        This method is a coroutine.
        """
        if self.receive_queue.empty() and self.transport.is_closing():
            return None
        return await self.receive_queue.get()

    def send(self, cmd):
        """Send a command to the SenseME fan."""
        self.transport.write(cmd.encode("utf-8"))


class SensemeProtocol(asyncio.Protocol):
    """Protocol for SenseME communication."""

    def __init__(self, name, endpoint: SensemeEndpoint):
        """Initialize Senseme Protocol."""
        self._name = name
        self._endpoint = endpoint

    # Protocol methods
    def connection_made(self, transport: asyncio.Protocol):
        """Socket connect on SenseME Protocol."""
        _LOGGER.debug("%s: Connected", self._name)
        self._endpoint.transport = transport
        self._endpoint.opened = True

    def connection_lost(self, exc):  # pylint: disable=unused-argument
        """Lost connection SenseME Protocol."""
        _LOGGER.debug("%s: Connection lost", self._name)
        self._endpoint.close()  # half-closed connections are not permitted

    # Streaming Protocol methods
    def data_received(self, data: str) -> str:
        """UDP packet received on SenseME Protocol."""
        if data:
            msg = data.decode("utf-8")
            try:
                self._endpoint.receive_queue.put_nowait(msg)
            except asyncio.QueueFull:
                _LOGGER.error("%s: Receive queue full", self._name)

    def eof_received(self) -> bool:
        """EOF received on SenseME Protocol."""
        return False  # tell the transport to close itself


class SensemeFan:
    """SensemeFan Class."""

    def __init__(
        self,
        name: str,
        id: str,  # pylint: disable=redefined-builtin
        ip: str,
        model: str,
        refresh_minutes: int = 1,
    ):
        """Initialize SensemeFan Class."""
        self._name = name
        self.refresh_minutes = refresh_minutes
        self._id = id
        self._ip = ip
        self._model = model
        self._room_name = None
        self._fw_name = ""
        self._fw_version = None
        self._has_light = None
        self._is_running = False
        self._is_connected = False
        self._data = dict()
        self._endpoint = None
        self._listener_task = None
        self._updater_task = None
        self._error_count = 0
        self._leftover = ""
        self._callbacks = []
        self._first_update = False

    def __eq__(self, other: Any) -> bool:
        """Equals magic method."""
        try:
            ip_addr = ipaddress.ip_address(other)
            return ip_addr == ipaddress.ip_address(self._ip)
        except ValueError:
            pass
        if isinstance(other, str):
            return other in [self._name, self._room_name]
        if isinstance(other, SensemeFan):
            return self._id == other.id
        return NotImplemented

    def __hash__(self) -> int:
        """Hash magic method."""
        return hash(self._id)

    def __str__(self) -> str:
        """Return string representation of SensemeFan object."""
        string = f"Name: {self._name}"
        if self._room_name is not None:
            string += f", Room Name: {self._room_name}"
        string += f", ID: {self._id}"
        string += f", IP: {self._ip}"
        string += f", Model: {self.model}"
        if self._fw_version is not None:
            string += f", FW Version: {self._fw_version}"
        if self._has_light is not None:
            string += f", Has Light: {self._has_light}"
        return string

    @property
    def is_sec_info_complete(self) -> bool:
        """Return if all secondary information is complete."""
        if self._fw_version is None:
            return False
        if self._has_light is None:
            return False
        if self._room_name is None:
            return False
        return True

    async def _query_fan(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, cmd: str
    ) -> str:
        """Send command to SenseME fan and parses response.

        None is returned if response was 'ERROR;PARSE'
        """
        msg = f"<{self.name};{cmd};GET>"
        # _LOGGER.debug("Sent command: '%s'", msg)
        writer.write(msg.encode("utf-8"))
        leftover = ""
        while True:
            # socket will throw a timeout error and abort this function if
            # no proper response is received
            data = await reader.read(1048)
            data = data.decode("utf-8")
            data = leftover + data
            leftover = ""
            # The data received may have multiple parenthesized data
            # points. Split them and put each message onto the queue.
            # Convert "(msg1)(msg2)(msg3)" to "(msg1)|(msg2)|(msg3)"
            # then split on the '|'
            for rsp in data.replace(")(", ")|(").split("|"):
                if rsp[-1] != ")":
                    self._leftover = rsp
                else:
                    # _LOGGER.debug("Received response: '%s'" ,p)
                    # remove '(fan name' at the beginning and the ')' at
                    # the end of the string
                    _, result = rsp[:-1].split(";", 1)
                    values = result.split(";")
                    key = ";".join(values[:-1])
                    value = ";".join(values[-1:])
                    if key == "ERROR":
                        _LOGGER.error("%s: Sent bad command: %s", self._name, msg)
                    else:
                        if cmd in rsp:
                            # correct response, return the value
                            return value

    async def fill_out_sec_info(self) -> bool:
        """Retrieve secondary info from the SenseME fan directly.

        Secondary data is fw_version, has_light and room_name which can only be
        filled in after connecting with the fan. The secondary data is also populated
        when the fan is started and after the fan is connected.
        This method is a coroutine.
        """
        if self.is_sec_info_complete:
            # info is already complete
            return True
        try:
            reader, writer = await asyncio.open_connection(self.ip, PORT)
            self._fw_name = await self._query_fan(reader, writer, "FW;NAME")
            self._fw_version = await self._query_fan(
                reader, writer, f"FW;{self._fw_name}"
            )
            self._has_light = (
                await self._query_fan(reader, writer, "DEVICE;LIGHT")
            ).upper() == "PRESENT"
            self._room_name = await self._query_fan(reader, writer, "GROUP;LIST")
            return True
        except OSError:
            _LOGGER.debug(
                "%s: Failed to retrieve secondary information from fan\n%s",
                self._name,
                traceback.format_exc(),
            )
            return False
        finally:
            # close the socket
            writer.close()
            await writer.wait_closed()

    @property
    def name(self) -> str:
        """Return name of fan."""
        return self._name

    @property
    def id(self) -> str:
        """Return id of fan. Also known as MAC address."""
        return self._id

    @property
    def mac(self) -> str:
        """Return MAC address of fan. Also known as id."""
        return self._id

    @property
    def ip(self) -> str:
        """Return IP address of fan."""
        return self._ip

    @property
    def connected(self) -> bool:
        """Return True when fan is connected."""
        return self._is_connected

    @property
    def model(self) -> str:
        """Return Model of fan."""
        return FAN_MODEL_TYPES.get(self._model.upper(), self._model.upper())

    @classmethod
    def models(cls) -> list:
        """Return list of possible model names."""
        no_duplicates = []
        for model_name in FAN_MODEL_TYPES.values():
            if model_name not in no_duplicates:
                no_duplicates.append(model_name)
        return no_duplicates

    @property
    def fw_version(self) -> str:
        """Return the version of the firmware running on the SenseME fan."""
        return self._fw_version

    @property
    def has_light(self) -> bool:
        """Return True if the fan has an installed light."""
        return self._has_light

    @property
    def device_indicators(self) -> str:
        """Return True if the device LED indicator is enabled."""
        value = self._data.get("DEVICE;INDICATORS", None)
        if value:
            return value == "ON"
        return None

    @device_indicators.setter
    def device_indicators(self, value: bool):
        """Enable/disable the device LED indicator."""
        if value:
            state = "ON"
        else:
            state = "OFF"
        self._send_command(f"DEVICE;INDICATORS;{state}")

    @property
    def device_beeper(self) -> bool:
        """Return the device audible alert enabled state."""
        status = self._data.get("DEVICE;BEEPER", None)
        if status:
            return status == "ON"
        return None

    @device_beeper.setter
    def device_beeper(self, value: bool):
        """Enable/disable the device audible alert."""
        if value:
            state = "ON"
        else:
            state = "OFF"
        self._send_command(f"DEVICE;BEEPER;{state}")

    @property
    def network_ap_on(self) -> bool:
        """Return the wireless access point running state."""
        status = self._data.get("NW;AP;STATUS", None)
        if status:
            return status == "ON"
        return None

    @property
    def network_dhcp_on(self) -> bool:
        """Return the fan local DHCP service running state."""
        dhcp = self._data.get("NW;DHCP", None)
        if dhcp:
            return dhcp == "ON"
        return None

    @property
    def network_ip(self) -> str:
        """Return the network IP address of the fan.

        This IP address is reported by the SenseME fan and not necessarily the same IP
        address used to connect with the fan.
        """
        addresses = self._data.get("NW;PARAMS;ACTUAL", None)
        if addresses:
            addresses = addresses.split(";")
            return addresses[0]
        return None

    @property
    def network_gateway(self) -> str:
        """Return the network gateway address of the fan."""
        addresses = self._data.get("NW;PARAMS;ACTUAL", None)
        if addresses:
            addresses = addresses.split(";")
            return addresses[1]
        return None

    @property
    def network_subnetmask(self) -> str:
        """Return the network gateway address of the fan."""
        addresses = self._data.get("NW;PARAMS;ACTUAL", None)
        if addresses:
            addresses = addresses.split(";")
            return addresses[2]
        return None

    @property
    def network_ssid(self) -> str:
        """Return the wireless SSID the fan is connected to."""
        return self._data.get("NW;SSID", None)

    @property
    def network_token(self) -> str:
        """Return the network token of the fan."""
        return self._data.get("NW;TOKEN", None)

    @property
    def room_status(self) -> bool:
        """Return True if the fan is in a room."""
        room_name = self._data.get("GROUP;LIST", None)
        room_type = self._data.get("GROUP;ROOM;TYPE", None)
        if room_name and room_type:
            return room_name != "EMPTY" and room_type != "0"
        return None

    @property
    def room_name(self) -> str:
        """Return the room name of the fan.

        'EMPTY' is returned if not in a group.
        """
        return self._data.get("GROUP;LIST", None)

    @property
    def room_type(self) -> str:
        """Return the room type of the fan."""
        room_type = int(self._data.get("GROUP;ROOM;TYPE", None))
        if room_type:
            if room_type >= len(ROOM_TYPES):
                room_type = 0
            return ROOM_TYPES[room_type]
        return None

    @property
    def fan_on(self) -> bool:
        """Return True when fan is on at any speed."""
        state = self._data.get("FAN;PWR", None)
        if state:
            return state == "ON"
        return None

    @fan_on.setter
    def fan_on(self, state: bool):
        """Set the fan power state."""
        state = "ON" if state else "OFF"
        self._send_command(f"FAN;PWR;{state}")

    @property
    def fan_speed(self) -> int:
        """Return the fan speed."""
        speed = self._data.get("FAN;SPD;ACTUAL", None)
        if speed:
            return int(speed)
        return None

    @fan_speed.setter
    def fan_speed(self, speed: int):
        """Set the fan speed."""
        if speed < 0:
            speed = 0
        elif speed > self.fan_speed_max:
            speed = self.fan_speed_max
        self._send_command(f"FAN;SPD;SET;{speed}")

    @property
    def fan_speed_min(self) -> int:
        """Return the fan speed minimum."""
        min_speed = self._data.get("FAN;SPD;MIN", None)
        if min_speed:
            return int(min_speed)
        return None

    @property
    def fan_speed_max(self) -> int:
        """Return the fan speed maximum."""
        max_speed = self._data.get("FAN;SPD;MAX", None)
        if max_speed:
            return int(max_speed)
        return None

    @property
    def fan_speed_limits_room(self) -> Tuple:
        """Return a tuple of the min/max fan speeds the room is configured to support.

        A room can limit the minimum/maximum fan speed while keeping the same number of
        speed settings. On the Haiku by BAF application this setting can be found by
        clicking the room info button. There must be at least one fan added to a room.
        """
        raw = self._data.get("FAN;BOOKENDS", None)
        if raw is None:
            return None
        values = raw.split(";")
        if len(values) != 2:
            return None
        min_speed = int(values[0])
        max_speed = int(values[1])
        return min_speed, max_speed

    @fan_speed_limits_room.setter
    def fan_speed_limits_room(self, speeds: Tuple):
        """Set a tuple of the min/max fan speeds the room is configured to support."""
        if speeds[0] >= speeds[1]:
            _LOGGER.error("Min speed cannot exceed max speed")
            return
        self._send_command(f"FAN;BOOKENDS;SET;{speeds[0]};{speeds[1]}")

    @property
    def fan_dir(self) -> str:
        """Return the fan direction."""
        return self._data.get("FAN;DIR", None)

    @fan_dir.setter
    def fan_dir(self, direction: str):
        """Set the fan direction."""
        if direction not in DIRECTIONS:
            raise ValueError(
                f"{direction} is not a valid direction. Must be one of {DIRECTIONS}"
            )
        self._send_command(f"FAN;DIR;SET;{direction}")

    @property
    def fan_whoosh(self) -> bool:
        """Return True when fan whoosh mode is on."""
        state = self._data.get("FAN;WHOOSH;STATUS", None)
        if state:
            return state == "ON"
        return None

    @fan_whoosh.setter
    def fan_whoosh(self, state: bool):
        """Set the fan whoosh mode."""
        value = "ON" if state else "OFF"
        self._send_command(f"FAN;WHOOSH;{value}")

    @property
    def fan_autocomfort(self) -> str:
        """Get the auto comfort mode from the fan.

        'OFF' no automatic adjustment.
        'COOLING' increases fan speed as temp increases.
        'HEATING' means slow mixing of air while room is occupied and faster mix speeds
                  while room is not occupied.
        'FOLLOWTSTAT' means change between 'COOLING' and 'HEATING based on thermostat.
        """
        return self._data.get("SMARTMODE;STATE", None)

    @fan_autocomfort.setter
    def fan_autocomfort(self, state: str):
        """Set the fan auto comfort mode.

        'OFF' no automatic adjustment.
        'COOLING' increases fan speed as temp increases.
        'HEATING' means slow mixing of air while room is occupied and faster mix speeds
                  while room is not occupied.
        'FOLLOWTSTAT' means change between 'COOLING' and 'HEATING based on thermostat.
        """
        value = "ON" if state else "OFF"
        self._send_command(f"SMARTMODE;STATE;SET;{value}")

    @property
    def fan_smartmode(self) -> str:
        """Get the current smart mode from the fan.

        'OFF' no automatic adjustment.
        'COOLING' increases fan speed as temp increases.
        'HEATING' means slow mixing of air while room is occupied and faster mix speeds
                  while room is not occupied.
        """
        return self._data.get("SMARTMODE;ACTUAL", None)

    @property
    def fan_cooltemp(self) -> float:
        """Return the auto shutoff temperature for 'COOLING' smart mode in Celsius."""
        temp = int(self._data.get("LEARN;ZEROTEMP", None))
        if temp:
            return float(temp) / 100.0
        return None

    @fan_cooltemp.setter
    def fan_cooltemp(self, temp: float):
        """Set the auto shutoff temperature for 'COOLING' smart mode in Celsius."""
        # force temperature into range
        if temp < 10:
            temp = 10
        elif temp > 31.5:
            temp = 31.5
        temp = int(round(temp * 100))
        self._send_command(f"LEARN;ZEROTEMP;SET;{temp}")

    @property
    def light_on(self) -> bool:
        """Return True when light is on at any brightness."""
        state = self._data.get("LIGHT;PWR", None)
        if state:
            return state == "ON"
        return None

    @light_on.setter
    def light_on(self, state: bool):
        """Set the light power state."""
        value = "ON" if state else "OFF"
        self._send_command(f"LIGHT;PWR;{value}")

    @property
    def light_brightness(self) -> int:
        """Return the light brightness."""
        level = self._data.get("LIGHT;LEVEL;ACTUAL", None)
        if level:
            return int(level)
        return None

    @light_brightness.setter
    def light_brightness(self, level: int):
        """Set the light brightness."""
        if level < 0:
            level = 0
        if level > 16:
            level = 16
        self._send_command(f"LIGHT;LEVEL;SET;{level}")

    @property
    def light_brightness_min(self) -> int:
        """Return the light brightness minimum."""
        min_brightness = self._data.get("LIGHT;LEVEL;MIN", None)
        if min_brightness:
            return int(min_brightness)
        return None

    @property
    def light_brightness_max(self) -> int:
        """Return the light brightness maximum."""
        max_brightness = self._data.get("LIGHT;LEVEL;MAX", None)
        if max_brightness:
            return int(max_brightness)
        return None

    @property
    def light_brightness_limits_room(self) -> Tuple:
        """Return a tuple of the min and max light brightness for the room.

        A room can limit the minimum/maximum light brightness while keeping the same
        number of light brightness levels. On the Haiku by BAF application this setting
        can be found by clicking the room info button. You have to have at least one
        fan with installed light added to a room.
        """
        raw = self._data.get("LIGHT;BOOKENDS", None)
        if raw is None:
            return None
        values = raw.split(";")
        if len(values) != 2:
            return None
        min_bright = int(values[0])
        max_bright = int(values[1])
        return min_bright, max_bright

    @light_brightness_limits_room.setter
    def light_brightness_limits_room(self, speeds: Tuple):
        """Set a tuple of the min and max light brightness for the room."""
        if speeds[0] >= speeds[1]:
            _LOGGER.error("Min speed cannot exceed max speed")
            return
        self._send_command(f"LIGHT;BOOKENDS;SET;{speeds[0]};{speeds[1]}")

    @property
    def motion_sensor(self) -> bool:
        """Return True when fan motion sensor says room is occupied.

        Available on all SenseME fans.
        """
        status = self._data.get("SNSROCC;STATUS", None)
        if status:
            return status == "OCCUPIED"
        return None

    @property
    def motion_fan_auto(self) -> bool:
        """Return True when fan is in automatic on with motion mode."""
        state = self._data.get("FAN;AUTO", None)
        if state:
            return state == "ON"
        return None

    @motion_fan_auto.setter
    def motion_fan_auto(self, state: bool):
        """Set the fan automatic on with motion mode."""
        state = "ON" if state else "OFF"
        self._send_command(f";FAN;AUTO;{state}")

    @property
    def motion_light_auto(self) -> bool:
        """Return True when light is in automatic on with motion mode."""
        state = self._data.get(";LIGHT;AUTO", None)
        if state:
            return state == "ON"
        return None

    @motion_light_auto.setter
    def motion_light_auto(self, state: bool):
        """Set the light automatic on with motion mode."""
        state = "ON" if state else "OFF"
        self._send_command(f";LIGHT;AUTO;{state}")

    def add_callback(self, callback: Callable):
        """Add callback function/coroutine. Called when parameters are updated."""
        if callback not in self._callbacks:
            self._callbacks.append(callback)
            if inspect.iscoroutinefunction(callback):
                _LOGGER.debug("%s: Added coroutine callback", self._name)
            else:
                _LOGGER.debug("%s: Added function callback", self._name)

    def remove_callback(self, callback):
        """Remove existing callback function/coroutine."""
        if callback in self._callbacks:
            self._callbacks.remove(callback)
            _LOGGER.debug("%s: Removed callback", self._name)

    async def update(self) -> bool:
        """Wait for first update of all parameters in SenseME fan.

        This method is a coroutine.
        """
        if self._first_update:
            return True
        if self._is_running is False:
            self.start()
        start = int(time.time())
        while not self._first_update:
            await asyncio.sleep(0.01)
            if int(time.time()) - start >= 5:
                return False
        return True

    def _execute_callbacks(self):
        """Run all callbacks to indicate something has changed."""
        for callback in self._callbacks:
            if inspect.iscoroutinefunction(callback):
                loop = asyncio.get_event_loop()
                loop.create_task(callback())
            else:
                callback()

    def _send_command(self, cmd):
        """Send a command to SenseME fan."""
        msg = f"<{self._name};{cmd}>"
        self._endpoint.send(msg)
        # _LOGGER.debug("%s: Command sent '%s'", self._name, cmd)

    def _update_status(self):
        """Issues commands to get complete status from fan."""
        _LOGGER.debug("%s: Status update", self._name)
        self._send_command("GETALL")
        # GETALL doesn't return the status of the motion detector,
        # so also request the motion detector status
        self._send_command("SNSROCC;STATUS;GET")

    async def _updater(self):
        """Periodically update fan parameters.

        This method is a coroutine.
        """
        while True:
            try:
                self._update_status()
                await asyncio.sleep(self.refresh_minutes * 60 + random.uniform(-10, 10))
            except asyncio.CancelledError:
                _LOGGER.debug("%s: Updater task cancelled", self._name)
                return
            except OSError:
                _LOGGER.debug(
                    "%s: Updater task error\n%s", self._name, traceback.format_exc()
                )
                await asyncio.sleep(self.refresh_minutes * 60 + random.uniform(-10, 10))
            except Exception:
                _LOGGER.error(
                    "%s: Unhandled updater task error\n%s",
                    self._name,
                    traceback.format_exc(),
                )
                await asyncio.sleep(10)
                raise
        _LOGGER.error("%s: Updater task ended", self._name)

    async def _listener(self):
        """Task that listens for fan status changes.

        This method is a coroutine.
        """
        while True:
            try:
                if self._error_count > 10:
                    _LOGGER.error("%s: Listener task too many errors", self._name)
                    self._is_connected = False
                    self._execute_callbacks()
                    self._updater_task.cancel()
                    if self._endpoint is not None:
                        self._endpoint.close()
                    self._endpoint = None
                    break
                if self._endpoint is None:
                    self._is_connected = False
                    self._execute_callbacks()
                    self._endpoint = SensemeEndpoint()
                    loop = asyncio.get_event_loop()
                    try:
                        _LOGGER.debug("%s: Connecting", self._name)
                        await loop.create_connection(
                            lambda: SensemeProtocol(self._name, self._endpoint),
                            self._ip,
                            PORT,
                        )
                    except OSError:
                        _LOGGER.debug(
                            "%s: Connect failed, " "try again in a minute\n%s",
                            self._name,
                            traceback.format_exc(),
                        )
                        self._endpoint = None
                        await asyncio.sleep(60)
                        continue
                    self._updater_task = loop.create_task(self._updater())
                    self._error_count = 0
                    self._is_connected = True
                    self._execute_callbacks()
                data = await self._endpoint.receive()
                if data is None:
                    # endpoint is closed, let task know it's time open another
                    _LOGGER.warning("%s: Connection lost", self._name)
                    self._is_connected = False
                    self._execute_callbacks()  # tell callbacks we disconnected
                    self._endpoint = None
                    self._updater_task.cancel()
                    await asyncio.sleep(1)
                    continue
                # add previous partial data to new data
                data = self._leftover + data
                self._leftover = ""
                # The data received may have multiple parenthesized data
                # points. Split them and put each individual message onto the
                # queue. Convert "(msg1)(msg2)(msg3)" to "(msg1)|(msg2)|(msg3)"
                # then split on the '|'
                for msg in data.replace(")(", ")|(").split("|"):
                    if msg[-1] != ")":
                        self._leftover = msg
                        continue
                    # remove '(fan name' at the beginning and the ')'
                    # at the end of the string
                    _, result = msg[:-1].split(";", 1)
                    # most messages have only one value at the end
                    valuecount = 1
                    if "BOOKENDS" in result:
                        valuecount = 2
                    elif "NW;PARAMS;ACTUAL" in result:
                        valuecount = 3
                    # split on ';' and the associate the correct
                    # number of values
                    values = result.split(";")
                    key = ";".join(values[:-valuecount])
                    value = ";".join(values[-valuecount:])
                    if key == "ERROR":
                        _LOGGER.error(
                            "%s: Command error response: '%s'", self._name, value,
                        )
                        continue
                    if key == "TIME;VALUE":
                        # ignore time parameter
                        continue
                    if self._data.get(key, "?????") == value:
                        # parameter is the same value
                        continue
                    if key == "WINTERMODE;STATE":
                        # first update complete when WINTERMODE;STATE is received
                        self._first_update = True
                    self._data[key] = value
                    _LOGGER.debug(
                        "%s: Param updated: [%s]='%s'", self._name, key, value,
                    )
                    # update certain local variables that are not part of data
                    if key == "FW;NAME":
                        self._fw_name = value
                    elif key == ("FW;" + self._fw_name):
                        self._fw_version = value
                    elif key == "DEVICE;LIGHT":
                        value = value.upper()
                        self._has_light = value == "PRESENT"
                    elif key == "GROUP;LIST":
                        self._room_name = value
                    self._execute_callbacks()
            except asyncio.CancelledError:
                _LOGGER.debug("%s: Listener task cancelled", self._name)
                return
            except OSError:
                _LOGGER.debug(
                    "%s: Listener task\n%s", self._name, traceback.format_exc()
                )
                self._error_count += 1
                await asyncio.sleep(1)
            except Exception:
                _LOGGER.error(
                    "%s: Listener task error\n%s", self._name, traceback.format_exc()
                )
                raise
        _LOGGER.error("%s: Listener task ended", self._name)

    def start(self):
        """Start the async task to handle responses from the fan."""
        if not self._is_running:
            loop = asyncio.get_event_loop()
            self._listener_task = loop.create_task(self._listener())
            self._is_running = True
            # _LOGGER.debug("%s: Started", self._name)

    def stop(self):
        """Signals thread to stop and returns immediately."""
        if self._is_running is True:
            self._listener_task.cancel()
            self._updater_task.cancel()
            self._is_running = False
